:- include('utils.pl').
:- include('parser.pl').
:- include('queries.pl').
:- include('pprint.pl').

% instantiate(PNode, feature(), fact)

run(File) :-
    parse(File, FSML),
    clear, extract,
    instantiate(FSML, Tree),
    ast(Tree, S), pprint(S).

testInstantiate(Tree) :-
    parse('examples/applet.fml', FSML), clear, extract, instantiate(FSML, Tree).

instantiate(start(_, Body), Tree) :- instantiate(Body, Tree).

% the first (declaration) FSML is instantiated 
instantiate(body(_, [D|_]), Tree) :- instantiate(D, Tree).

% should have parent only if main FSML definition
instantiate(declaration(D), Tree) :- instantiate(empty, D, ResTree),
    isValidPred(ResTree), second(ResTree, Tree).

instantiate(PNode, featureDef(NameExType, FeatureBody), Tree) :-
    instantiate(PNode, FeatureBody, ResTree),
    just(ResTree, [featureDef, NameExType], Tree).

% featureBody definition
instantiate(PNode, featureBody(CodeMapping, Subfeatures, Constraints, Local),
           Tree) :-
    instantiate(PNode,
    featureBody(CodeMapping, Subfeatures, Constraints, Local),
                card(false, 0, *), Tree).

instantiate(PNode, subfeature(Subfeature, Card, Constraints), Tree) :-
    instantiate(PNode, Subfeature, Card, ResTree),
    just(ResTree, [subfeature, Card, Constraints], Tree).

% featureBody: local definitions ignored
instantiate(PNode, featureBody(CodeMapping, Subfeatures, Constraints, _),
           Card, Tree) :-
    instantiate(PNode, CodeMapping, Subfeatures, Card, ResTree),
    just(ResTree, [featureBody, Constraints], Tree).

instantiate(PNode, [], fact(Name, _, _, Node, _), Tree) :-
    isParent(PNode, Node), !,
    Tree = maybe(ok, mapping(Name, Node)).

instantiate(PNode, [subfeatures(Subfeatures)], fact(Name, _, _, Node, _), Tree) :-
    isParent(PNode, Node), !,
    map(instantiate(Node), Subfeatures, ResTrees),
    justList(ResTrees, [mapping, Name, Node], Tree).

instantiate(PNode, inlineFeatureDef(NameExType, _, FeatureBody), Card, Tree) :-
    instantiate(PNode, FeatureBody, Card, ResTree),
    just(ResTree, [inlineFeatureDef, NameExType], Tree).

% featureBody: no code mapping
instantiate(PNode, [], [subfeatures(Subfeatures)], Card, Tree) :-
    map(instantiate(PNode), Subfeatures, Subtrees),
    filterMapValidate(Subtrees, Card, Tree).

% featureBody: with code mapping, CM type ignored
instantiate(PNode, [pattern(_, PatName, Params)], Subfeatures, Card, Tree) :-
    append([query, PatName | Params], [Fact], ParamsFact),
    Query =.. ParamsFact,
    findall(Fact,
        (Query, Fact = fact(_, Node, _, _, _), isParent(PNode, Node)), FS),
    map(instantiate(PNode, Subfeatures), FS, Subtrees),
    filterMapValidate(Subtrees, Card, Tree).

% ugly...
filterMapValidate(Subtrees, Card, maybe(Res,
    featureBodyPattern(FilteredSubtrees))) :-
    filter(isValidPred, Subtrees, ValidSubtrees),
    map(second, ValidSubtrees, FilteredSubtrees),
    isValid(Card, FilteredSubtrees, Res).

isParent(PNode, PNode) :- !.
isParent(PNode, Node) :-
    fact(_, PNode1, _, Node, _), !,
    isParent(PNode, PNode1).

just(maybe(Res, Tree), Out, maybe(Res, OutTree)) :-
    append(Out, [Tree], OutT),
    OutTree =.. OutT.

justList(ResTrees, Out, ResOutTree) :-
    isValidPredList(ResTrees, Res),
    map(second, ResTrees, Trees),
    just(maybe(Res, Trees), Out, ResOutTree).

isValidPredList([], ok) :- !.
isValidPredList([maybe(ok, _)|XS], Res) :- isValidPredList(XS, Res), !.
isValidPredList(_, invalid).

second(maybe(_, S), S).

isValidPred(maybe(ok, _)).
%isValidPred(maybe(missing, _)).

isValidBasic(Card, XS) :- isValid(Card, XS, ok), !.
isValidBasic(Card, XS) :- isValid(Card, XS, missing).

isValid(card(true, Min, Max), XS, ok) :- isValidRange(Min, Max, XS), !.
isValid(card(_, Min, Max), XS, ok) :- isValidRange(Min, Max, XS), !.
%isValid(card(false, _, _), _, missing) :- !.
isValid(card(false, _, _), _, ok) :- !.
isValid(card(_, _, _), _, invalid).

isValidRange(Min, Max, XS) :- isValidMin(Min, XS), isValidMax(Max, XS).
        
isValidMin(0, _) :- !.
isValidMin(1, [_|_]) :- !.
isValidMin(L, XS) :- isValidNum(L, XS, =<).

isValidMax(*, _) :- !.
isValidMax(L, XS) :- isValidNum(L, XS, >=).

isValidNum(L, XS, REL) :- length(XS, LXS), RELOP =.. [REL, L, LXS], RELOP.