%:- include('simple.pl').
%:- include('utils.pl').

% extracts facts from AST and adds to database
% fact(name, parentPtr, myType, myPtr, myInfo)

clear :- abolish(fact/5), reset_gensym(node).
extract :- start(X), extract(start(X)).
show :- listing(fact).

extract(start(Projects)) :-
%    addFact(fact(start, PNode, start, Node, empty)),
    mapio(extract(empty), Projects).
extract(PNode, project(Name, Classes)) :-
    addFact(fact(project, PNode, jproject, Node, empty)),
    addFact(fact(projectName, Node, string, _, Name)),
    mapio(extract(Node), Classes).
extract(PNode, class(Name, BaseClass, Interfaces, Methods)) :-
    addFact(fact(class, PNode, jclass, Node, empty)),
    addFact(fact(assignableTo, Node, string, _, BaseClass)),
    addFact(fact(fullyQualifiedName, Node, string, _, Name)),
    mapio(extract(Node, implements, string), Interfaces),
    mapio(extract(Node), Methods).
extract(PNode, method(Name, RetType, Args, Body)) :-
    addFact(fact(hasMethod, PNode, jmethod, Node, empty)),
    addFact(fact(methodName, Node, string, _, Name)),
    addFact(fact(methodRetType, Node, jtype, _, RetType)),
    map(toTypeVal, Args, TVArgs),
    enumArgs(TVArgs, EArgs),
    mapio(extract(Node, argVal), EArgs),
    mapio(extract(Node), Body).
extract(PNode, methodCall(Object, Name, Args)) :-
    addFact(fact(calls, PNode, jmethodcall, Node, empty)),
    addFact(fact(methodObject, Node, string, _, Object)),
    addFact(fact(methodName, Node, string, _, Name)),
    enumArgs(Args, EArgs),
    mapio(extract(Node, argVal), EArgs).
extract(PNode, Name, Type, Element) :-
    Fact =.. [fact, Name, PNode, Type, _, Element],
    addFact(Fact).
extract(PNode, Name, (N, (Type, Val))) :-
    Fact =.. [fact, Name, PNode, Type, _, (N, Val)],
    addFact(Fact).

addFact(fact(Name, PNode, Type, Node, Info)) :-
    gensym(node, Node),
    assert(fact(Name, PNode, Type, Node, Info)).

enumArgs([], []) :- !.
enumArgs(Args, EArgs) :-
    length(Args, L),
    numlist(1, L, NS),
    map(number_atom, NS, ANS),
    zip(ANS, Args, EArgs).

toTypeVal(X, (X, [])).