% output configuration
:- set_prolog_flag(toplevel_print_options,
    [quoted(true), portray(true), max_depth(0), attributes(portray)]).

number_atom(N, A) :- atom_number(A, N).

% applies partially evaluated F to each element
map(_, [], []) :- !.
map(PF, [H|T], [NH|NT]) :-
    applyFun(PF, [H, NH]),
    map(PF, T, NT).

% applies partially evaluated F to each element but F does not return result
mapio(_, []) :- !.
mapio(PF, [H|T]) :-
    applyFun(PF, [H]),
    mapio(PF, T).

% tries to satisfy and return given goal
satisfy(X, X) :- X.

% zips two lists
zip([], _, []) :- !.
zip(_, [], []) :- !.
zip([X|XS], [Y|YS], [(X, Y)|Tail]) :- zip(XS, YS, Tail).

% replicates element
replicate(0, _, []) :- !.
replicate(N, X, [X|XS]) :- NP is N - 1, replicate(NP, X, XS).

% Kleene star for DCG
gcdStar(_, [])       --> [].
gcdStar(Fun, [X|XS]) --> callFun(Fun, [X]), gcdStar(Fun, XS).

gcdPlus(Fun, [X|XS]) --> gcdStar(Fun, [X|XS]).

% [separator exp]+
gcdStarSep(Separator, Fun, [X|XS]) -->
    Separator, callFun(Fun, [X]), gcdStarSep(Separator, Fun, XS).
gcdStarSep(_, _, [])               --> [].

% [separator exp]*
gcdStar(Separator, Fun, [X|XS]) -->
    callFun(Fun, [X]), gcdStarSep(Separator, Fun, XS).
gcdStar(_, _, [])               --> [].

% [separator(S) exp]+
gcdStarSep(Separator, S, Fun, [S, X|XS]) -->
    callFun(Separator, [S]), callFun(Fun, [X]), gcdStarSep(Separator, _, Fun, XS).
gcdStarSep(_, _, _, [])                  --> [].

% [separator(S) exp]*
gcdStar(Separator, _, Fun, [X|XS]) -->
    callFun(Fun, [X]), gcdStarSep(Separator, _, Fun, XS).
gcdStar(_, _, _, [])               --> [].

gcdPlus(Separator, Fun, [X|XS]) --> gcdStar(Separator, Fun, [X|XS]).

gcdPlus(Separator, S, Fun, [X|XS]) --> gcdStar(Separator, S, Fun, [X|XS]).

applyFun(PF, Params) :-
    PF =.. [F|OldParams],
    append(OldParams, Params, NewParams),
    apply(F, NewParams).

%optional(Fun, X) --> callFun(Fun, X), !.
%optional(_, [[]])  --> [].

optional(Fun, [X]) --> callFun(Fun, [X]), !.
optional(_, [])    --> [].

callFun(Fun, X, In, Out) :-
    append(X, [In, Out], Y),
    applyFun(Fun, Y).
callFun(Fun, In, Out)    :- callFun(Fun, [], In, Out).

takeWhile(Pred, [X|XS], [X|Tail], Out) :-
    applyFun(Pred, [X]),
    takeWhile(Pred, XS, Tail, Out), !.
takeWhile(_, [], X, X) :- !.

dropWhile(Pred, [H|Tail], Out) :-
    applyFun(Pred, [H]),
    dropWhile(Pred, Tail, Out), !.
dropWhile(_, X, X).

init([_|[]], []).
init([X|XS], [X|Init]) :- init(XS, Init).

filter(_, [], []) :- !.
filter(Pred, [X|XS], [X|YS]) :- applyFun(Pred, [X]), filter(Pred, XS, YS), !.
filter(Pred, [_|XS], YS) :- filter(Pred, XS, YS).
